@@ -1,7 +1,7 @@
 /*
  * Copyright (C) Igor Sysoev
  * Copyright (C) 2007 Manlio Perillo (manlio.perillo@gmail.com)
- * Copyright (C) 2010-2013 Phusion
+ * Copyright (C) 2010-2015 Phusion
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -31,12 +31,15 @@
 
 #include <sys/types.h>
 #include <pwd.h>
+#include <stdlib.h>
+#include <assert.h>
 
 #include "ngx_http_passenger_module.h"
 #include "Configuration.h"
 #include "ContentHandler.h"
 #include "common/Constants.h"
-#include "common/agents/LoggingAgent/FilterSupport.h"
+#include "common/agent/UstRouter/FilterSupport.h"
+#include "common/Utils/modp_b64.h"
 
 
 static ngx_str_t headers_to_hide[] = {
@@ -56,17 +59,24 @@ static ngx_path_init_t  ngx_http_proxy_temp_path = {
     ngx_string(NGX_HTTP_PROXY_TEMP_PATH), { 1, 2, 0 }
 };
 
+static ngx_int_t merge_headers(ngx_conf_t *cf, passenger_loc_conf_t *conf,
+    passenger_loc_conf_t *prev);
+static ngx_int_t merge_string_array(ngx_conf_t *cf, ngx_array_t **prev,
+    ngx_array_t **conf);
+static ngx_int_t merge_string_keyval_table(ngx_conf_t *cf, ngx_array_t **prev,
+    ngx_array_t **conf);
+
 
 void *
 passenger_create_main_conf(ngx_conf_t *cf)
 {
     passenger_main_conf_t *conf;
-    
+
     conf = ngx_pcalloc(cf->pool, sizeof(passenger_main_conf_t));
     if (conf == NULL) {
         return NGX_CONF_ERROR;
     }
-    
+
     conf->ctl = ngx_array_create(cf->pool, 1, sizeof(ngx_keyval_t));
     if (conf->ctl == NULL) {
         return NGX_CONF_ERROR;
@@ -74,12 +84,22 @@ passenger_create_main_conf(ngx_conf_t *cf)
     conf->default_ruby.data = NULL;
     conf->default_ruby.len = 0;
     conf->log_level = (ngx_int_t) NGX_CONF_UNSET;
-    conf->debug_log_file.data = NULL;
-    conf->debug_log_file.len = 0;
+    conf->log_file.data = NULL;
+    conf->log_file.len = 0;
+    conf->file_descriptor_log_file.data = NULL;
+    conf->file_descriptor_log_file.len = 0;
+    conf->data_buffer_dir.data = NULL;
+    conf->data_buffer_dir.len = 0;
+    conf->instance_registry_dir.data = NULL;
+    conf->instance_registry_dir.len = 0;
     conf->abort_on_startup_error = NGX_CONF_UNSET;
-    conf->max_pool_size = (ngx_uint_t) NGX_CONF_UNSET;
-    conf->pool_idle_time = (ngx_uint_t) NGX_CONF_UNSET;
+    conf->max_pool_size = NGX_CONF_UNSET_UINT;
+    conf->pool_idle_time = NGX_CONF_UNSET_UINT;
+    conf->response_buffer_high_watermark = NGX_CONF_UNSET_UINT;
+    conf->stat_throttle_rate = NGX_CONF_UNSET_UINT;
     conf->user_switching = NGX_CONF_UNSET;
+    conf->show_version_in_header = NGX_CONF_UNSET;
+    conf->turbocaching = NGX_CONF_UNSET;
     conf->default_user.data = NULL;
     conf->default_user.len  = 0;
     conf->default_group.data = NULL;
@@ -95,12 +115,12 @@ passenger_create_main_conf(ngx_conf_t *cf)
     conf->union_station_gateway_cert.len = 0;
     conf->union_station_proxy_address.data = NULL;
     conf->union_station_proxy_address.len = 0;
-    
+
     conf->prestart_uris = ngx_array_create(cf->pool, 1, sizeof(ngx_str_t));
     if (conf->prestart_uris == NULL) {
         return NGX_CONF_ERROR;
     }
-    
+
     return conf;
 }
 
@@ -111,10 +131,10 @@ passenger_init_main_conf(ngx_conf_t *cf, void *conf_pointer)
     struct passwd         *user_entry;
     struct group          *group_entry;
     char buf[128];
-    
+
     conf = &passenger_main_conf;
     *conf = *((passenger_main_conf_t *) conf_pointer);
-    
+
     if (conf->default_ruby.len == 0) {
         conf->default_ruby.data = (u_char *) DEFAULT_RUBY;
         conf->default_ruby.len = strlen(DEFAULT_RUBY);
@@ -123,27 +143,55 @@ passenger_init_main_conf(ngx_conf_t *cf, void *conf_pointer)
     if (conf->log_level == (ngx_int_t) NGX_CONF_UNSET) {
         conf->log_level = DEFAULT_LOG_LEVEL;
     }
-    
-    if (conf->debug_log_file.len == 0) {
-        conf->debug_log_file.data = (u_char *) "";
+
+    if (conf->log_file.len == 0) {
+        conf->log_file.data = (u_char *) "";
     }
-    
+
+    if (conf->file_descriptor_log_file.len == 0) {
+        conf->file_descriptor_log_file.data = (u_char *) "";
+    }
+
+    if (conf->data_buffer_dir.len == 0) {
+        conf->data_buffer_dir.data = (u_char *) "";
+    }
+
+    if (conf->instance_registry_dir.len == 0) {
+        conf->instance_registry_dir.data = (u_char *) "";
+    }
+
     if (conf->abort_on_startup_error == NGX_CONF_UNSET) {
         conf->abort_on_startup_error = 0;
     }
-    
-    if (conf->max_pool_size == (ngx_uint_t) NGX_CONF_UNSET) {
+
+    if (conf->max_pool_size == NGX_CONF_UNSET_UINT) {
         conf->max_pool_size = DEFAULT_MAX_POOL_SIZE;
     }
-    
-    if (conf->pool_idle_time == (ngx_uint_t) NGX_CONF_UNSET) {
+
+    if (conf->pool_idle_time == NGX_CONF_UNSET_UINT) {
         conf->pool_idle_time = DEFAULT_POOL_IDLE_TIME;
     }
-    
+
+    if (conf->response_buffer_high_watermark == NGX_CONF_UNSET_UINT) {
+        conf->response_buffer_high_watermark = DEFAULT_RESPONSE_BUFFER_HIGH_WATERMARK;
+    }
+
+    if (conf->stat_throttle_rate == NGX_CONF_UNSET_UINT) {
+        conf->stat_throttle_rate = DEFAULT_STAT_THROTTLE_RATE;
+    }
+
     if (conf->user_switching == NGX_CONF_UNSET) {
         conf->user_switching = 1;
     }
-    
+
+    if (conf->show_version_in_header == NGX_CONF_UNSET) {
+        conf->show_version_in_header = 1;
+    }
+
+    if (conf->turbocaching == NGX_CONF_UNSET) {
+        conf->turbocaching = 1;
+    }
+
     if (conf->default_user.len == 0) {
         conf->default_user.len  = sizeof(DEFAULT_WEB_APP_USER) - 1;
         conf->default_user.data = (u_char *) DEFAULT_WEB_APP_USER;
@@ -157,7 +205,7 @@ passenger_init_main_conf(ngx_conf_t *cf, void *conf_pointer)
     if (user_entry == NULL) {
         return "The user specified by the 'default_user' option does not exist.";
     }
-    
+
     if (conf->default_group.len > 0) {
         if (conf->default_group.len > sizeof(buf) - 1) {
             return "Value for 'default_group' is too long.";
@@ -169,42 +217,43 @@ passenger_init_main_conf(ngx_conf_t *cf, void *conf_pointer)
             return "The group specified by the 'default_group' option does not exist.";
         }
     }
-    
+
     if (conf->analytics_log_user.len == 0) {
         conf->analytics_log_user.len  = sizeof(DEFAULT_ANALYTICS_LOG_USER) - 1;
         conf->analytics_log_user.data = (u_char *) DEFAULT_ANALYTICS_LOG_USER;
     }
-    
+
     if (conf->analytics_log_group.len == 0) {
         conf->analytics_log_group.len  = sizeof(DEFAULT_ANALYTICS_LOG_GROUP) - 1;
         conf->analytics_log_group.data = (u_char *) DEFAULT_ANALYTICS_LOG_GROUP;
     }
-    
+
     if (conf->union_station_gateway_address.len == 0) {
         conf->union_station_gateway_address.len = sizeof(DEFAULT_UNION_STATION_GATEWAY_ADDRESS) - 1;
         conf->union_station_gateway_address.data = (u_char *) DEFAULT_UNION_STATION_GATEWAY_ADDRESS;
     }
-    
+
     if (conf->union_station_gateway_port == (ngx_uint_t) NGX_CONF_UNSET) {
         conf->union_station_gateway_port = DEFAULT_UNION_STATION_GATEWAY_PORT;
     }
-    
+
     if (conf->union_station_gateway_cert.len == 0) {
         conf->union_station_gateway_cert.data = (u_char *) "";
     }
-    
+
     if (conf->union_station_proxy_address.len == 0) {
         conf->union_station_proxy_address.data = (u_char *) "";
     }
-    
+
     return NGX_CONF_OK;
 }
 
+#include "CreateLocationConfig.c"
+
 void *
 passenger_create_loc_conf(ngx_conf_t *cf)
 {
     passenger_loc_conf_t  *conf;
-    ngx_keyval_t          *kv;
 
     conf = ngx_pcalloc(cf->pool, sizeof(passenger_loc_conf_t));
     if (conf == NULL) {
@@ -226,7 +275,7 @@ passenger_create_loc_conf(ngx_conf_t *cf)
      *     conf->upstream_config.store_values = NULL;
      */
 
-    #include "CreateLocationConfig.c"
+    generated_set_conf_part(conf);
 
     /******************************/
     /******************************/
@@ -236,17 +285,29 @@ passenger_create_loc_conf(ngx_conf_t *cf)
 
     conf->upstream_config.store = NGX_CONF_UNSET;
     conf->upstream_config.store_access = NGX_CONF_UNSET_UINT;
+    #if NGINX_VERSION_NUM >= 1007005
+        conf->upstream_config.next_upstream_tries = NGX_CONF_UNSET_UINT;
+    #endif
     conf->upstream_config.buffering = NGX_CONF_UNSET;
     conf->upstream_config.ignore_client_abort = NGX_CONF_UNSET;
+    #if NGINX_VERSION_NUM >= 1007007
+        conf->upstream_config.force_ranges = NGX_CONF_UNSET;
+    #endif
 
     conf->upstream_config.local = NGX_CONF_UNSET_PTR;
 
     conf->upstream_config.connect_timeout = NGX_CONF_UNSET_MSEC;
     conf->upstream_config.send_timeout = NGX_CONF_UNSET_MSEC;
     conf->upstream_config.read_timeout = NGX_CONF_UNSET_MSEC;
+    #if NGINX_VERSION_NUM >= 1007005
+        conf->upstream_config.next_upstream_timeout = NGX_CONF_UNSET_MSEC;
+    #endif
 
     conf->upstream_config.send_lowat = NGX_CONF_UNSET_SIZE;
     conf->upstream_config.buffer_size = NGX_CONF_UNSET_SIZE;
+    #if NGINX_VERSION_NUM >= 1007007
+        conf->upstream_config.limit_rate = NGX_CONF_UNSET_SIZE;
+    #endif
 
     conf->upstream_config.busy_buffers_size_conf = NGX_CONF_UNSET_SIZE;
     conf->upstream_config.max_temp_file_size_conf = NGX_CONF_UNSET_SIZE;
@@ -257,9 +318,9 @@ passenger_create_loc_conf(ngx_conf_t *cf)
 
 #if (NGX_HTTP_CACHE)
     #if NGINX_VERSION_NUM >= 1007009
-        conf->upstream_config.cache = NGX_CONF_UNSET_PTR;
-    #else
         conf->upstream_config.cache = NGX_CONF_UNSET;
+    #else
+        conf->upstream_config.cache = NGX_CONF_UNSET_PTR;
     #endif
     conf->upstream_config.cache_min_uses = NGX_CONF_UNSET_UINT;
     conf->upstream_config.cache_bypass = NGX_CONF_UNSET_PTR;
@@ -270,102 +331,126 @@ passenger_create_loc_conf(ngx_conf_t *cf)
     #if NGINX_VERSION_NUM >= 1007008
         conf->upstream_config.cache_lock_age = NGX_CONF_UNSET_MSEC;
     #endif
+    #if NGINX_VERSION_NUM >= 1006000
+        conf->upstream_config.cache_revalidate = NGX_CONF_UNSET;
+    #endif
 #endif
 
     conf->upstream_config.intercept_errors = NGX_CONF_UNSET;
 
     conf->upstream_config.cyclic_temp_file = 0;
     conf->upstream_config.change_buffering = 1;
-    
-    #define DEFINE_VAR_TO_PASS(header_name, var_name) \
-        kv = ngx_array_push(conf->vars_source);       \
-        kv->key.data = (u_char *) header_name;        \
-        kv->key.len  = strlen(header_name) + 1;       \
-        kv->value.data = (u_char *) var_name;         \
-        kv->value.len  = strlen(var_name) + 1
-    
-    conf->vars_source = ngx_array_create(cf->pool, 4, sizeof(ngx_keyval_t));
-    if (conf->vars_source == NULL) {
-        return NGX_CONF_ERROR;
-    }
-    
-    DEFINE_VAR_TO_PASS("SCGI",            "1");
-    DEFINE_VAR_TO_PASS("QUERY_STRING",    "$query_string");
-    DEFINE_VAR_TO_PASS("REQUEST_METHOD",  "$request_method");
-    DEFINE_VAR_TO_PASS("SERVER_PROTOCOL", "$server_protocol");
-    DEFINE_VAR_TO_PASS("SERVER_SOFTWARE", "nginx/$nginx_version");
-    DEFINE_VAR_TO_PASS("REMOTE_ADDR",     "$remote_addr");
-    DEFINE_VAR_TO_PASS("REMOTE_PORT",     "$remote_port");
-    DEFINE_VAR_TO_PASS("SERVER_ADDR",     "$server_addr");
-    DEFINE_VAR_TO_PASS("SERVER_PORT",     "$server_port");
 
     ngx_str_set(&conf->upstream_config.module, "passenger");
 
+    conf->options_cache.data  = NULL;
+    conf->options_cache.len   = 0;
+    conf->env_vars_cache.data = NULL;
+    conf->env_vars_cache.len  = 0;
+
     return conf;
 }
 
-static void
+#include "CacheLocationConfig.c"
+
+static ngx_int_t
 cache_loc_conf_options(ngx_conf_t *cf, passenger_loc_conf_t *conf)
 {
-    #include "CacheLocationConfig.c"
+    ngx_uint_t     i;
+    ngx_keyval_t  *env_vars;
+    size_t         unencoded_len;
+    u_char        *unencoded_buf;
+
+    if (generated_cache_location_part(cf, conf) == 0) {
+    	return NGX_ERROR;
+    }
+
+    if (conf->env_vars != NULL) {
+    	size_t len = 0;
+    	u_char *buf;
+    	u_char *pos;
+
+        /* Cache env vars data as base64-serialized string.
+         * First, calculate the length of the unencoded data.
+         */
+
+        unencoded_len = 0;
+        env_vars = (ngx_keyval_t *) conf->env_vars->elts;
+
+        for (i = 0; i < conf->env_vars->nelts; i++) {
+            unencoded_len += env_vars[i].key.len + 1 + env_vars[i].value.len + 1;
+        }
+
+        /* Create the unecoded data. */
+
+        unencoded_buf = pos = (u_char *) malloc(unencoded_len);
+        if (unencoded_buf == NULL) {
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "cannot allocate buffer of %z bytes for environment variables data",
+                               unencoded_len);
+            return NGX_ERROR;
+        }
+
+        for (i = 0; i < conf->env_vars->nelts; i++) {
+            pos = ngx_copy(pos, env_vars[i].key.data, env_vars[i].key.len);
+            *pos = '\0';
+            pos++;
+
+            pos = ngx_copy(pos, env_vars[i].value.data, env_vars[i].value.len);
+            *pos = '\0';
+            pos++;
+        }
+
+        assert((size_t) (pos - unencoded_buf) == unencoded_len);
+
+        /* Create base64-serialized string. */
+
+        buf = ngx_palloc(cf->pool, modp_b64_encode_len(unencoded_len));
+        if (buf == NULL) {
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "cannot allocate buffer of %z bytes for base64 encoding",
+                               modp_b64_encode_len(unencoded_len));
+            return NGX_ERROR;
+        }
+        len = modp_b64_encode((char *) buf, (const char *) unencoded_buf, unencoded_len);
+        if (len == (size_t) -1) {
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "error during base64 encoding");
+            free(unencoded_buf);
+            return NGX_ERROR;
+        }
+
+        conf->env_vars_cache.data = buf;
+        conf->env_vars_cache.len = len;
+        free(unencoded_buf);
+    }
+
+    return NGX_OK;
 }
 
+#include "MergeLocationConfig.c"
+
 char *
 passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
 {
     passenger_loc_conf_t         *prev = parent;
     passenger_loc_conf_t         *conf = child;
+    ngx_http_core_loc_conf_t     *clcf;
 
-    u_char                       *p;
     size_t                        size;
-    uintptr_t                    *code;
-    ngx_uint_t                    i;
-    ngx_str_t                    *prev_base_uris, *base_uri;
-    ngx_str_t                    *prev_union_station_filters, *union_station_filter;
-    ngx_keyval_t                 *src;
     ngx_hash_init_t               hash;
-    ngx_http_script_compile_t     sc;
-    ngx_http_script_copy_code_t  *copy;
 
-    #include "MergeLocationConfig.c"
+    clcf = ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);
+
+    if (generated_merge_part(conf, prev, cf) == 0) {
+    	return NGX_CONF_ERROR;
+    }
+
     if (prev->options_cache.data == NULL) {
-        cache_loc_conf_options(cf, prev);
-    }
-    cache_loc_conf_options(cf, conf);
-    
-    if (prev->base_uris != NGX_CONF_UNSET_PTR) {
-        if (conf->base_uris == NGX_CONF_UNSET_PTR) {
-            conf->base_uris = ngx_array_create(cf->pool, 4, sizeof(ngx_str_t));
-            if (conf->base_uris == NULL) {
-                return NGX_CONF_ERROR;
-            }
-        }
-        
-        prev_base_uris = (ngx_str_t *) prev->base_uris->elts;
-        for (i = 0; i < prev->base_uris->nelts; i++) {
-            base_uri = (ngx_str_t *) ngx_array_push(conf->base_uris);
-            if (base_uri == NULL) {
-                return NGX_CONF_ERROR;
-            }
-            *base_uri = prev_base_uris[i];
-        }
-    }
-    
-    if (prev->union_station_filters != NGX_CONF_UNSET_PTR) {
-        if (conf->union_station_filters == NGX_CONF_UNSET_PTR) {
-            conf->union_station_filters = ngx_array_create(cf->pool, 4, sizeof(ngx_str_t));
-            if (conf->union_station_filters == NULL) {
-                return NGX_CONF_ERROR;
-            }
-        }
-        
-        prev_union_station_filters = (ngx_str_t *) prev->union_station_filters->elts;
-        for (i = 0; i < prev->union_station_filters->nelts; i++) {
-            union_station_filter = (ngx_str_t *) ngx_array_push(conf->union_station_filters);
-            if (union_station_filter == NULL) {
-                return NGX_CONF_ERROR;
-            }
-            *union_station_filter = prev_union_station_filters[i];
+        if (cache_loc_conf_options(cf, prev) != NGX_OK) {
+            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                               "cannot create " PROGRAM_NAME " configuration cache");
+            return NGX_CONF_ERROR;
         }
     }
 
@@ -404,12 +489,22 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_conf_merge_uint_value(conf->upstream_config.store_access,
                               prev->upstream_config.store_access, 0600);
 
+    #if NGINX_VERSION_NUM >= 1007005
+        ngx_conf_merge_uint_value(conf->upstream_config.next_upstream_tries,
+                                  prev->upstream_config.next_upstream_tries, 0);
+    #endif
+
     ngx_conf_merge_value(conf->upstream_config.buffering,
                          prev->upstream_config.buffering, 0);
 
     ngx_conf_merge_value(conf->upstream_config.ignore_client_abort,
                          prev->upstream_config.ignore_client_abort, 0);
 
+    #if NGINX_VERSION_NUM >= 1007007
+        ngx_conf_merge_value(conf->upstream_config.force_ranges,
+                             prev->upstream_config.force_ranges, 0);
+    #endif
+
     ngx_conf_merge_ptr_value(conf->upstream_config.local,
                              prev->upstream_config.local, NULL);
 
@@ -422,6 +517,11 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_conf_merge_msec_value(conf->upstream_config.read_timeout,
                               prev->upstream_config.read_timeout, 12000000);
 
+    #if NGINX_VERSION_NUM >= 1007005
+        ngx_conf_merge_msec_value(conf->upstream_config.next_upstream_timeout,
+                                  prev->upstream_config.next_upstream_timeout, 0);
+    #endif
+
     ngx_conf_merge_size_value(conf->upstream_config.send_lowat,
                               prev->upstream_config.send_lowat, 0);
 
@@ -429,6 +529,11 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
                               prev->upstream_config.buffer_size,
                               16 * 1024);
 
+    #if NGINX_VERSION_NUM >= 1007007
+        ngx_conf_merge_size_value(conf->upstream_config.limit_rate,
+                                  prev->upstream_config.limit_rate, 0);
+    #endif
+
 
     ngx_conf_merge_bufs_value(conf->upstream_config.bufs, prev->upstream_config.bufs,
                               8, 16 * 1024);
@@ -580,11 +685,6 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
         }
     #endif
 
-    #if NGINX_VERSION_NUM >= 1007008
-        ngx_conf_merge_msec_value(conf->upstream_config.cache_lock_age,
-                                  prev->upstream_config.cache_lock_age, 5000);
-    #endif
-
     ngx_conf_merge_uint_value(conf->upstream_config.cache_min_uses,
                               prev->upstream_config.cache_min_uses, 1);
 
@@ -627,6 +727,19 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_conf_merge_msec_value(conf->upstream_config.cache_lock_timeout,
                               prev->upstream_config.cache_lock_timeout, 5000);
 
+    ngx_conf_merge_value(conf->upstream_config.cache_revalidate,
+                         prev->upstream_config.cache_revalidate, 0);
+
+    #if NGINX_VERSION_NUM >= 1007008
+        ngx_conf_merge_msec_value(conf->upstream_config.cache_lock_age,
+                                  prev->upstream_config.cache_lock_age, 5000);
+    #endif
+
+    #if NGINX_VERSION_NUM >= 1006000
+        ngx_conf_merge_value(conf->upstream_config.cache_revalidate,
+                             prev->upstream_config.cache_revalidate, 0);
+    #endif
+
 #endif
 
     ngx_conf_merge_value(conf->upstream_config.pass_request_headers,
@@ -653,97 +766,189 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
         conf->upstream_config.upstream = prev->upstream_config.upstream;
     }
 
-    if (conf->vars_source == NULL) {
-        conf->flushes = prev->flushes;
-        conf->vars_len = prev->vars_len;
-        conf->vars = prev->vars;
-        conf->vars_source = prev->vars_source;
+    if (conf->enabled == 1 /* and not NGX_CONF_UNSET */
+     && passenger_main_conf.root_dir.len != 0
+     && clcf->handler == NULL /* no handler set by other modules */)
+    {
+        clcf->handler = passenger_content_handler;
+    }
 
-        if (conf->vars_source == NULL) {
-            return NGX_CONF_OK;
+    conf->headers_hash_bucket_size = ngx_align(conf->headers_hash_bucket_size,
+                                               ngx_cacheline_size);
+    hash.max_size = conf->headers_hash_max_size;
+    hash.bucket_size = conf->headers_hash_bucket_size;
+    hash.name = "passenger_headers_hash";
+
+    if (merge_headers(cf, conf, prev) != NGX_OK) {
+        return NGX_CONF_ERROR;
+    }
+
+    if (cache_loc_conf_options(cf, conf) != NGX_OK) {
+        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
+                           "cannot create " PROGRAM_NAME " configuration cache");
+        return NGX_CONF_ERROR;
+    }
+
+    return NGX_CONF_OK;
+}
+
+static ngx_int_t
+merge_headers(ngx_conf_t *cf, passenger_loc_conf_t *conf, passenger_loc_conf_t *prev)
+{
+    u_char                       *p;
+    size_t                        size;
+    uintptr_t                    *code;
+    ngx_uint_t                    i;
+    ngx_array_t                   headers_names, headers_merged;
+    ngx_keyval_t                 *src, *s;
+    ngx_hash_key_t               *hk;
+    ngx_hash_init_t               hash;
+    ngx_http_script_compile_t     sc;
+    ngx_http_script_copy_code_t  *copy;
+
+    if (conf->headers_source == NULL) {
+        conf->flushes = prev->flushes;
+        conf->headers_set_len = prev->headers_set_len;
+        conf->headers_set = prev->headers_set;
+        conf->headers_set_hash = prev->headers_set_hash;
+        conf->headers_source = prev->headers_source;
+    }
+
+    if (conf->headers_set_hash.buckets
+#if (NGX_HTTP_CACHE)
+    #if NGINX_VERSION_NUM >= 1007009
+        && ((conf->upstream_config.cache == NGX_CONF_UNSET) == (prev->upstream_config.cache == NGX_CONF_UNSET))
+    #else
+        && ((conf->upstream_config.cache == NGX_CONF_UNSET_PTR) == (prev->upstream_config.cache == NGX_CONF_UNSET_PTR))
+    #endif
+#endif
+       )
+    {
+        return NGX_OK;
+    }
+
+
+    if (ngx_array_init(&headers_names, cf->temp_pool, 4, sizeof(ngx_hash_key_t))
+        != NGX_OK)
+    {
+        return NGX_ERROR;
+    }
+
+    if (ngx_array_init(&headers_merged, cf->temp_pool, 4, sizeof(ngx_keyval_t))
+        != NGX_OK)
+    {
+        return NGX_ERROR;
+    }
+
+    if (conf->headers_source == NULL) {
+        conf->headers_source = ngx_array_create(cf->pool, 4,
+                                                sizeof(ngx_keyval_t));
+        if (conf->headers_source == NULL) {
+            return NGX_ERROR;
         }
     }
 
-    conf->vars_len = ngx_array_create(cf->pool, 64, 1);
-    if (conf->vars_len == NULL) {
-        return NGX_CONF_ERROR;
+    conf->headers_set_len = ngx_array_create(cf->pool, 64, 1);
+    if (conf->headers_set_len == NULL) {
+        return NGX_ERROR;
     }
 
-    conf->vars = ngx_array_create(cf->pool, 512, 1);
-    if (conf->vars == NULL) {
-        return NGX_CONF_ERROR;
+    conf->headers_set = ngx_array_create(cf->pool, 512, 1);
+    if (conf->headers_set == NULL) {
+        return NGX_ERROR;
     }
 
-    src = conf->vars_source->elts;
-    for (i = 0; i < conf->vars_source->nelts; i++) {
+
+    src = conf->headers_source->elts;
+    for (i = 0; i < conf->headers_source->nelts; i++) {
+
+        s = ngx_array_push(&headers_merged);
+        if (s == NULL) {
+            return NGX_ERROR;
+        }
+
+        *s = src[i];
+    }
+
+
+    src = headers_merged.elts;
+    for (i = 0; i < headers_merged.nelts; i++) {
+
+        hk = ngx_array_push(&headers_names);
+        if (hk == NULL) {
+            return NGX_ERROR;
+        }
+
+        hk->key = src[i].key;
+        hk->key_hash = ngx_hash_key_lc(src[i].key.data, src[i].key.len);
+        hk->value = (void *) 1;
+
+        if (src[i].value.len == 0) {
+            continue;
+        }
 
         if (ngx_http_script_variables_count(&src[i].value) == 0) {
-            copy = ngx_array_push_n(conf->vars_len,
+            copy = ngx_array_push_n(conf->headers_set_len,
                                     sizeof(ngx_http_script_copy_code_t));
             if (copy == NULL) {
-                return NGX_CONF_ERROR;
-            }
-
-            copy->code = (ngx_http_script_code_pt)
-                                                  ngx_http_script_copy_len_code;
-            copy->len = src[i].key.len;
-
-
-            copy = ngx_array_push_n(conf->vars_len,
-                                    sizeof(ngx_http_script_copy_code_t));
-            if (copy == NULL) {
-                return NGX_CONF_ERROR;
+                return NGX_ERROR;
             }
 
             copy->code = (ngx_http_script_code_pt)
                                                  ngx_http_script_copy_len_code;
-            copy->len = src[i].value.len;
+            copy->len = src[i].key.len + sizeof(": ") - 1
+                        + src[i].value.len + sizeof(CRLF) - 1;
 
 
             size = (sizeof(ngx_http_script_copy_code_t)
-                       + src[i].key.len + src[i].value.len
+                       + src[i].key.len + sizeof(": ") - 1
+                       + src[i].value.len + sizeof(CRLF) - 1
                        + sizeof(uintptr_t) - 1)
                     & ~(sizeof(uintptr_t) - 1);
 
-            copy = ngx_array_push_n(conf->vars, size);
+            copy = ngx_array_push_n(conf->headers_set, size);
             if (copy == NULL) {
-                return NGX_CONF_ERROR;
+                return NGX_ERROR;
             }
 
             copy->code = ngx_http_script_copy_code;
-            copy->len = src[i].key.len + src[i].value.len;
+            copy->len = src[i].key.len + sizeof(": ") - 1
+                        + src[i].value.len + sizeof(CRLF) - 1;
 
             p = (u_char *) copy + sizeof(ngx_http_script_copy_code_t);
 
             p = ngx_cpymem(p, src[i].key.data, src[i].key.len);
-            ngx_memcpy(p, src[i].value.data, src[i].value.len);
+            *p++ = ':'; *p++ = ' ';
+            p = ngx_cpymem(p, src[i].value.data, src[i].value.len);
+            *p++ = CR; *p = LF;
 
         } else {
-            copy = ngx_array_push_n(conf->vars_len,
+            copy = ngx_array_push_n(conf->headers_set_len,
                                     sizeof(ngx_http_script_copy_code_t));
             if (copy == NULL) {
-                return NGX_CONF_ERROR;
+                return NGX_ERROR;
             }
 
             copy->code = (ngx_http_script_code_pt)
                                                  ngx_http_script_copy_len_code;
-            copy->len = src[i].key.len;
+            copy->len = src[i].key.len + sizeof(": ") - 1;
 
 
             size = (sizeof(ngx_http_script_copy_code_t)
-                    + src[i].key.len + sizeof(uintptr_t) - 1)
+                    + src[i].key.len + sizeof(": ") - 1 + sizeof(uintptr_t) - 1)
                     & ~(sizeof(uintptr_t) - 1);
 
-            copy = ngx_array_push_n(conf->vars, size);
+            copy = ngx_array_push_n(conf->headers_set, size);
             if (copy == NULL) {
-                return NGX_CONF_ERROR;
+                return NGX_ERROR;
             }
 
             copy->code = ngx_http_script_copy_code;
-            copy->len = src[i].key.len;
+            copy->len = src[i].key.len + sizeof(": ") - 1;
 
             p = (u_char *) copy + sizeof(ngx_http_script_copy_code_t);
-            ngx_memcpy(p, src[i].key.data, src[i].key.len);
+            p = ngx_cpymem(p, src[i].key.data, src[i].key.len);
+            *p++ = ':'; *p = ' ';
 
 
             ngx_memzero(&sc, sizeof(ngx_http_script_compile_t));
@@ -751,45 +956,154 @@ passenger_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
             sc.cf = cf;
             sc.source = &src[i].value;
             sc.flushes = &conf->flushes;
-            sc.lengths = &conf->vars_len;
-            sc.values = &conf->vars;
+            sc.lengths = &conf->headers_set_len;
+            sc.values = &conf->headers_set;
 
             if (ngx_http_script_compile(&sc) != NGX_OK) {
-                return NGX_CONF_ERROR;
+                return NGX_ERROR;
             }
+
+
+            copy = ngx_array_push_n(conf->headers_set_len,
+                                    sizeof(ngx_http_script_copy_code_t));
+            if (copy == NULL) {
+                return NGX_ERROR;
+            }
+
+            copy->code = (ngx_http_script_code_pt)
+                                                 ngx_http_script_copy_len_code;
+            copy->len = sizeof(CRLF) - 1;
+
+
+            size = (sizeof(ngx_http_script_copy_code_t)
+                    + sizeof(CRLF) - 1 + sizeof(uintptr_t) - 1)
+                    & ~(sizeof(uintptr_t) - 1);
+
+            copy = ngx_array_push_n(conf->headers_set, size);
+            if (copy == NULL) {
+                return NGX_ERROR;
+            }
+
+            copy->code = ngx_http_script_copy_code;
+            copy->len = sizeof(CRLF) - 1;
+
+            p = (u_char *) copy + sizeof(ngx_http_script_copy_code_t);
+            *p++ = CR; *p = LF;
         }
 
-        code = ngx_array_push_n(conf->vars_len, sizeof(uintptr_t));
+        code = ngx_array_push_n(conf->headers_set_len, sizeof(uintptr_t));
         if (code == NULL) {
-            return NGX_CONF_ERROR;
+            return NGX_ERROR;
         }
 
         *code = (uintptr_t) NULL;
 
-
-        code = ngx_array_push_n(conf->vars, sizeof(uintptr_t));
+        code = ngx_array_push_n(conf->headers_set, sizeof(uintptr_t));
         if (code == NULL) {
-            return NGX_CONF_ERROR;
+            return NGX_ERROR;
         }
 
         *code = (uintptr_t) NULL;
     }
 
-    code = ngx_array_push_n(conf->vars_len, sizeof(uintptr_t));
+    code = ngx_array_push_n(conf->headers_set_len, sizeof(uintptr_t));
     if (code == NULL) {
-        return NGX_CONF_ERROR;
+        return NGX_ERROR;
     }
 
     *code = (uintptr_t) NULL;
 
-    return NGX_CONF_OK;
+
+    hash.hash = &conf->headers_set_hash;
+    hash.key = ngx_hash_key_lc;
+    hash.max_size = conf->headers_hash_max_size;
+    hash.bucket_size = conf->headers_hash_bucket_size;
+    hash.name = "passenger_headers_hash";
+    hash.pool = cf->pool;
+    hash.temp_pool = NULL;
+
+    return ngx_hash_init(&hash, headers_names.elts, headers_names.nelts);
+}
+
+static ngx_int_t
+merge_string_array(ngx_conf_t *cf, ngx_array_t **prev, ngx_array_t **conf)
+{
+    ngx_str_t  *prev_elems, *elem;
+    ngx_uint_t  i;
+
+    if (*prev != NGX_CONF_UNSET_PTR) {
+        if (*conf == NGX_CONF_UNSET_PTR) {
+            *conf = ngx_array_create(cf->pool, 4, sizeof(ngx_str_t));
+            if (*conf == NULL) {
+                return NGX_ERROR;
+            }
+        }
+
+        prev_elems = (ngx_str_t *) (*prev)->elts;
+        for (i = 0; i < (*prev)->nelts; i++) {
+            elem = (ngx_str_t *) ngx_array_push(*conf);
+            if (elem == NULL) {
+                return NGX_ERROR;
+            }
+            *elem = prev_elems[i];
+        }
+    }
+
+    return NGX_OK;
+}
+
+static int
+string_keyval_has_key(ngx_array_t *table, ngx_str_t *key)
+{
+    ngx_keyval_t  *elems;
+    ngx_uint_t     i;
+
+    elems = (ngx_keyval_t *) table->elts;
+    for (i = 0; i < table->nelts; i++) {
+        if (elems[i].key.len == key->len
+         && memcmp(elems[i].key.data, key->data, key->len) == 0)
+        {
+            return 1;
+        }
+    }
+
+    return 0;
+}
+
+static ngx_int_t
+merge_string_keyval_table(ngx_conf_t *cf, ngx_array_t **prev, ngx_array_t **conf)
+{
+    ngx_keyval_t  *prev_elems, *elem;
+    ngx_uint_t     i;
+
+    if (*prev != NULL) {
+        if (*conf == NULL) {
+            *conf = ngx_array_create(cf->pool, 4, sizeof(ngx_keyval_t));
+            if (*conf == NULL) {
+                return NGX_ERROR;
+            }
+        }
+
+        prev_elems = (ngx_keyval_t *) (*prev)->elts;
+        for (i = 0; i < (*prev)->nelts; i++) {
+            if (!string_keyval_has_key(*conf, &prev_elems[i].key)) {
+                elem = (ngx_keyval_t *) ngx_array_push(*conf);
+                if (elem == NULL) {
+                    return NGX_ERROR;
+                }
+                *elem = prev_elems[i];
+            }
+        }
+    }
+
+    return NGX_OK;
 }
 
 #ifndef PASSENGER_IS_ENTERPRISE
 static char *
 passenger_enterprise_only(ngx_conf_t *cf, ngx_command_t *cmd, void *conf) {
     return ": this feature is only available in Phusion Passenger Enterprise. "
-        "You are currently running the open source Phusion Passenger Enterprise. "
+        "You are currently running the open source Phusion Passenger. "
         "Please learn more about and/or buy Phusion Passenger Enterprise at https://www.phusionpassenger.com/enterprise ;";
 }
 #endif
@@ -805,11 +1119,11 @@ passenger_enabled(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
     value = cf->args->elts;
     if (ngx_strcasecmp(value[1].data, (u_char *) "on") == 0) {
         passenger_conf->enabled = 1;
-        
+
         /* Register a placeholder value as upstream address. The real upstream
-         * address (the helper agent socket filename) will be set while processing
-         * requests, because we can't start the helper agent until config
-         * loading is done.
+         * address (the Passenger core socket filename) will be set while processing
+         * requests, because we can't start the watchdog (and thus the Passenger core)
+         * until config loading is done.
          */
         ngx_memzero(&upstream_url, sizeof(ngx_url_t));
         upstream_url.url = pp_placeholder_upstream_address;
@@ -818,7 +1132,7 @@ passenger_enabled(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
         if (passenger_conf->upstream_config.upstream == NULL) {
             return NGX_CONF_ERROR;
         }
-        
+
         clcf = ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);
         clcf->handler = passenger_content_handler;
 
@@ -932,7 +1246,7 @@ set_null_terminated_keyval_slot(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
     kv->key.len  = value[1].len + 1;
     last = ngx_copy(kv->key.data, value[1].data, value[1].len);
     *last = '\0';
-    
+
     kv->value.data = ngx_palloc(cf->pool, value[2].len + 1);
     kv->value.len  = value[2].len + 1;
     last = ngx_copy(kv->value.data, value[2].data, value[2].len);
@@ -979,18 +1293,32 @@ const ngx_command_t passenger_commands[] = {
       offsetof(passenger_main_conf_t, log_level),
       NULL },
 
-    { ngx_string("passenger_debug_log_file"),
+    { ngx_string("passenger_log_file"),
       NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
       NGX_HTTP_MAIN_CONF_OFFSET,
-      offsetof(passenger_main_conf_t, debug_log_file),
+      offsetof(passenger_main_conf_t, log_file),
       NULL },
 
-    { ngx_string("passenger_temp_dir"),
+    { ngx_string("passenger_file_descriptor_log_file"),
       NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
       NGX_HTTP_MAIN_CONF_OFFSET,
-      offsetof(passenger_main_conf_t, temp_dir),
+      offsetof(passenger_main_conf_t, file_descriptor_log_file),
+      NULL },
+
+    { ngx_string("passenger_data_buffer_dir"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_str_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, data_buffer_dir),
+      NULL },
+
+    { ngx_string("passenger_instance_registry_dir"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_str_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, instance_registry_dir),
       NULL },
 
     { ngx_string("passenger_pre_start"),
@@ -1021,6 +1349,34 @@ const ngx_command_t passenger_commands[] = {
       offsetof(passenger_main_conf_t, pool_idle_time),
       NULL },
 
+    { ngx_string("passenger_response_buffer_high_watermark"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_num_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, response_buffer_high_watermark),
+      NULL },
+
+    { ngx_string("passenger_stat_throttle_rate"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_num_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, stat_throttle_rate),
+      NULL },
+
+    { ngx_string("passenger_show_version_in_header"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, show_version_in_header),
+      NULL },
+
+    { ngx_string("passenger_turbocaching"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, turbocaching),
+      NULL },
+
     { ngx_string("passenger_user_switching"),
       NGX_HTTP_MAIN_CONF | NGX_CONF_FLAG,
       ngx_conf_set_flag_slot,
@@ -1056,6 +1412,13 @@ const ngx_command_t passenger_commands[] = {
       offsetof(passenger_main_conf_t, analytics_log_group),
       NULL },
 
+    { ngx_string("passenger_read_timeout"),
+      NGX_HTTP_MAIN_CONF | NGX_HTTP_SRV_CONF | NGX_HTTP_LOC_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_msec_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(passenger_loc_conf_t, upstream_config.read_timeout),
+      NULL },
+
     { ngx_string("union_station_gateway_address"),
       NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
       ngx_conf_set_str_slot,
@@ -1088,6 +1451,15 @@ const ngx_command_t passenger_commands[] = {
 
     #include "ConfigurationCommands.c"
 
+    /******** Backward compatibility ********/
+
+    { ngx_string("passenger_debug_log_file"),
+      NGX_HTTP_MAIN_CONF | NGX_CONF_TAKE1,
+      ngx_conf_set_str_slot,
+      NGX_HTTP_MAIN_CONF_OFFSET,
+      offsetof(passenger_main_conf_t, log_file),
+      NULL },
+
     ngx_null_command
 };
 
